/**
	AD9951-628v6.cpp

	Firmware pour piloter un circuit DDS : AD9951.
	par Silicium628.

	Ce fichier source "AD9951.cpp" est libre, vous pouvez le redistribuer et/ou le modifier selon
	les termes de la Licence Publique Générale GNU.
**/


#include <stdio.h>
#include <stdint.h> 
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "AD9951-628v6.h"

AD9951::AD9951()
{
	FTW = 0;
	OSC_REFCLK = 0;
}


void AD9951::set_ref_clk( uint64_t F_CLK_i, uint8_t REFCLK_Multiplier_i)
{
// cette fonction doit être appellée au moins une fois avant celle "calcul_FTW"
	OSC_REFCLK = F_CLK_i * REFCLK_Multiplier_i; // la fréquence d'horloge entrée sur le pin 9 (OSC_REFCLK)
}



void AD9951::reset()
{
/*
Active High Hardware Reset Pin. Assertion of the RESET pin forces the AD9951 to the initial state,
as described in the I/O port register map
Attention :  ne pas faire un RESET après l'init, ce qui effacerait le paramètrage
*/	
 _delay_ms(1);
  port_AD9951 |=  pin_RST;
 _delay_ms(10);
  port_AD9951 &= ~pin_RST; // l'opérateur "~" donne le complement (inverse des bits). ne pas confondre avec l'opérateur "!"
 _delay_ms(10);
}



void AD9951::impulse_clk()
{
 _delay_us(5);  // temps necessaire à la stabilisation du niveau envoyé
  port_AD9951 |=  pin_SCLK;
 _delay_us(5);
  port_AD9951 &= ~pin_SCLK; // l'operateur "~" donne le complement (inverse des bits). ne pas confondre avec l'opérateur "!"
 _delay_us(5);
}


void AD9951::impulse_IO_update()
{
// _delay_us(10);
  port_AD9951 |=  pin_io_update;
 _delay_us(10);
  port_AD9951 &= ~pin_io_update; // l'operateur "~" donne le complement (inverse des bits). ne pas confondre avec l'opérateur "!"
 _delay_us(10);
}



/***********************************************************************
f_out = FTW x SYSCLK / 2³² nous dit le datasheet p:12
2³² = 4294967296 exactement.
en en déduit :

FTW = f_out x 2³² / SYSCLK
// ici SYSCLK = 20MHz x 20 = 400 MHz = 4x10E8  // **** cas d'un TCXO 20MHz ****
// ici SYSCLK = 10MHz x 20 = 200 MHz = 2x10E8  // **** cas d'une synchro par GPS ****
ici SYSCLK = 12MHz x 20 = 240 MHz  = 240x10E6  // **** cas d'une synchro par GPS ****

//FTW = f_out x (2³² / 4E8) = f_out * (2³⁰ / 1E8) = 10.73741824 x f_out   **** cas d'un TCXO 20MHz ****
FTW = f_out x (2³² / 2E8) = 21.47483648 x f_out
Remarque : SYSCLK est aussi nommé OSC_REFCLK dans le datasheet lorsqu'il s'agit du AD9951 en particulier

************************************************************************/
uint8_t AD9951::calcul_FTW(uint32_t frequence, uint32_t *FTW_entier, uint8_t *FTW_decim)
{
	static int DIGITS=0;
// rappel : FTW et OSC_REFCLK sont des uint_64
	uint64_t y, FTW2;

	y = frequence * 4294967296;  // frequence * 2³²
	FTW = y / OSC_REFCLK; //  OSC_REFCLK (ici) = 12MHz * 20 = 240 x 10E+6
	*FTW_entier = FTW;

	if (frequence <= 42e+6)
	{
		FTW2 = y * 100; // ATTENTION : FTW2 est > 64 bits si frequence > 42MHz
		FTW2 /= OSC_REFCLK;
		*FTW_decim = FTW2 % 100; // modulo = reste de la division - donne les deux chiffres après la virgule
	}
	else
	{
		FTW2 = y * 10; 
		FTW2 /= OSC_REFCLK;
		*FTW_decim = 10 * (FTW2 % 10); // modulo = reste de la division - donne les un chiffre après la virgule
	}


	

/**
pour vérification du chiffre après la virgule...
calculs effectués avec "calc" en ligne de commande sous Linux (C-style arbitrary precision calculator (version 2.12.4.4))

		1*(2^32)/240E6 = 17.89  ok   -> 1Hz   1789 ? tiens, tiens...
Remarque : pour les fréquences très petites, la partie décimale perdue est du mmême ordre de grandeur que la partie entière,
donc la précision fréquencielle du DDS sans correction est très mauvaise pour les fréquences faibles.
La première fréquence ayant un FTW qui tombe "presque juste" est 115 Hz. (deux zéros après la virgule)
	115*(2^32)/240E6 = 2058.00516266666666666667
Vient apès 278 Hz.	

   1e3*(2^32)/240E6 = 17895.69 ok
   69 / 17895 = 3.85582564962279966471E-3
   
 2.4E6*(2^32)/240E6 = 42949672.96  ok
 
	12E6*(2^32)/240E06 =	214748364,80  ok  -> 12MHz
	80 / 214748364 = 0.0000003725290312340  = 0.37253e-6 -> valeur de la correcrion à apporter
	
	20E6*(2^32)/240E6 = 357913941.33 ok
	40E6*(2^32)/240E6 = 715827882.66 ok
	42E6*(2^32)/240E6 = 751619276.80  ok
	
	43E6*(2^32)/240E6 = 769514973.86  HS (100 * frequence * 4294967296  déborde sur 64 bits !)
	44E6*(2^32)/240E6 = 787410670.93  HS
donc la correction ne sera plus valable au dela de 43MHz sauf à trouver une astuce pour ne ma déborder	
  
	10*42E6*(2^32)/240E6  = 0b 1 11000000 00000000 00000000 00000000
	100*42E6*(2^32) = 0b   11111010 01010110 11101010 00000000 00000000 00000000 00000000 00000000
	100*43E6*(2^32) = 0b 1 00000000 01001100 11001011 00000000 00000000 00000000 00000000 00000000 (ce qui dépasse 64bits !)
**/


/**
// les deux lignes qui suivent permettent de "forcer manuellement" un (très léger) glissement de fréquence
	FTW += 128;
	FTW -= adj;
**/

/**
	 On introduit forcément une erreur d'arrondi, et le résultat est à +/- 1 digit
	 par exemple pour une fréquence de sortie de 10MHz le glissemment par rapport au 10MHz étalon est de 26s
	 pour une période de 100ns -> 3,8 E-9 soit 1 seconde/3ans ce qui n'est pas si mal !!
	 si l'on veut obtenir la précision de E-12 du GPS, il faut envisager d'utiliser autre chose qu'un synthétiseur DDS 32 bits.
	 
	 Pour des fréquences multiples ou sous-multiples de 10MHz,
	 ou dans un rapport fractionnaire simple, une simple PLL convient avec diviseur en logique câblée.
	 pour d'autres fréquences, il faut étudier la question de plus près.

	 Dans le cas de ce DDS, on peut faire en sorte que SYSCLK / 2³² tombe juste, c'est à dire que SYSCLK = n x 2³²
	 2³² étant lui même >> (4 milliards) on ne peut envisager que le cas n < 1 par puissances de 2, c.a.d n=1/2, 1/4, 1/8 ...
	 mais afin que SYSCLK soit < 400MHz, ce qui est la limite haute pour ce circuit, il faut diviser 2³² par une puissance de 2 supérieure à 10, donc au moins par 16.
	 ce qui donne  SYSCLK = 2³² / 2⁴ = 2²⁸ = 268435456 Hz soit une fréquence de quartz (compte tenu du *16 de la PLL interne) de 16777216 Hz. (= 2²⁴ =16 777 216 MHz)
	 La formule devient FTW = f_out x 2³²/2²⁸ = f_out * 16
	On peut obtenir la fréquence  16.777216 MHz en multipliant par 4 la fréquence d'un quartz de 4.194304MHz (par une PLL en HCMOS)
	remarque : on trouve facilement des quartz de 4.194304MHz (=2²²)
	 
**/

	return 0;
}



void AD9951::out(uint8_t octet_i)
{
// MSB first (= mode par defaut)
	uint8_t i;

//envoi de la frequence
  uint8_t masque; 

	for (i=0;  i < 8; i++)  
	{
		masque = 0b10000000 >> i; // le '1' se deplace de gauche a droite
		if ( (octet_i & masque) != 0) { port_AD9951 |=  pin_SDI; }	 else { port_AD9951 &= ~pin_SDI; }
		impulse_clk();
	}
		
}



void AD9951::DDS_init()
{
// Attention :  ne pas faire un RESET après cette init, ce qui effacerait le paramètrage
	uint8_t CFR1[4]; 
	uint8_t CFR2[3];
	
// ============================ REGISTRE CRF1 - 32 bits ===================================================================		
// Control Function Register No.1 (CFR1) (0x00)


//PREMIER OCTET :

// CFR1<0>: Not Used, Leave at 0 
// CFR1<1> SYNC_CLK Disable Bit;  CFR1<1> = 0 (default). The SYNC_CLK pin is active.
// CFR1<2>: Not Used
// CFR1<3>: External Power-Down Mode; CFR1<3> = 0 (default). The external power-down mode selected is the rapid recovery power-down mode. In this mode,
//when the PWRDWNCTL input pin is high, the digital logic  and the DAC digital logic are powered down. The DAC bias circuitry, PLL, oscillator, and clock input circuitry are not powered down.
// CFR1<4>: Clock Input Power-Down Bit; CFR1<4> = 0 (default). The clock input circuitry is enabled for operation. 
// CFR1<5>: DAC Power-Down Bit; 	CFR1<5> = 0 (default). The DAC is enabled for operation.
// CFR1<6>: Not Used
// CFR1<7>: Digital Power-Down Bit; CFR1<7> = 0 (default). All digital functions and clocks are active.

	CFR1[0]=0b00000000;

//DEUXIEME OCTET

// CFR1<8> = 0 (default). MSB first format is active.	
// CFR1<9>: SDIO Input Only; CFR1<9> = 1. The serial data I/O pin (SDIO) is configured as an input only pin (3-wire serial programming mode). (CE N'EST PAS LA VALEUR PAR DEFAUT)
// CFR1<10>: Clear Phase Accumulator; CFR1<10> = 0 (default). The phase accumulator functions as normal.
// CFR1<11>: Not Used
//CFR1<12>: Sine/Cosine Select Bit; CFR1<12> = 0 (default). The angle-to-amplitude conversion logic employs a COSINE function.
// CFR1<13>: Auto-Clear Phase Accumulator Bit; CFR1<13> = 0 (default), the current state of the phase accumulator remains unchanged when the frequency tuning word is applied.
// CFR1<14..15>: Not Used

	CFR1[1]=0b00000010; // CFR1<9>: SDIO Input Only;
		
//TROISIEME OCTET		
		
// CFR1<16..21>: Not Used
// CFR1<22>: Software Manual Synchronization of Multiple AD9951; CFR1<22> = 0 (default). The manual synchronization feature is inactive.
// CFR1<23>: Automatic Synchronization Enable Bit;  CFR1<23> = 0 (default). The automatic synchronization feature of multiple AD9951s is inactive. 
		
	CFR1[2]=0b00000000;
		
// CFR1<24>: Auto Shaped On-Off Keying Enable Bit (Only Valid when CFR1<25> Is Active High); CFR1<24> = 0 (default). When CFR1<25> is active, a Logic 0 on CFR1<24> enables the 
//manual shaped on-off keying operation. Each amplitude sample sent to the DAC is multiplied by the amplitude scale factor. See the Shaped On-Off Keying section for details	
// CFR1<25>: Shaped On-Off Keying Enable Bit; CFR1<25> = 0 (default). Shaped on-off keying is bypassed.
// CFR1<26>: Amplitude Ramp Rate Load Control Bit; CFR1<26> = 0 (default). The amplitude ramp rate timer is loaded only upon timeout (timer == 1) and is not loaded due to an I/O UPDATE input signal. 
// CFR1<31:27>: Not Used 

	CFR1[3]=0b00000000;

// ENVOI DE CETTE CONFIGURATION AU DDS : (voir page 22 : "Example Operation")

	IB= 0b00000000;
	out( IB); // instruction byte -> ordre d'ecrire dans le registre CRF1 (32 bits attendus)
	
	out( CFR1[3]);
	out( CFR1[2]) ;
	out( CFR1[1]);
	out( CFR1[0]);
	
	impulse_IO_update();
		
// ============================ REGISTRE CRF2 - 24bits ============ voir le pdf du composant =========================================================		
		
// CFR2<1:0>: Charge Pump Current Control Bits
// CFR2<2>: VCO Range Control Bit
/*
This bit is used to control the range setting on the VCO.
When CFR2<2> == 0 (default), the VCO operates in a range of
100 MHz to 250 MHz. When CFR2<2> == 1, the VCO operates
in a range of 250 MHz to 400 MHz.
*/
// CFR2<7:3>: Reference Clock Multiplier Control Bits
/*
This 5-bit word controls the multiplier value out of the clock-
multiplier (PLL) block. Valid values are decimal 4 to 20 (0x04 to
0x14). Values entered outside this range will bypass the clock
multiplier. See the Phase-Locked Loop (PLL) section for details
20 = 0b10100
*/		
// CFR2<8>: Not Used
// CFR2<9>: CRYSTAL OUT Enable Bit
// CFR2<10>: Hardware Manual Sync Enable Bit
// CFR2<11>: High Speed Sync Enable Bit
// CFR2<23:12>: Not Used

//CFR2[0] = 0b00100100; // bit2 =1 -> the VCO operates in a range of 250 MHz to 400 MHz;  bits7..3 = 0b00100 -> =  4  (REFCLK Multiplier) -> cas d'un Qx 100MHz
	CFR2[0] = 0b10100100; // bit2 =1 -> the VCO operates in a range of 250 MHz to 400 MHz;  bits7..3 = 0b10100 -> = 20  (REFCLK Multiplier) -> cas d'un Qx ou CTXO 20MHz
	
	CFR2[1] = 0b00000010; // bit CFR2<9>: CRYSTAL OUT Enable Bit -> =1
	CFR2[2] = 0b00000000;
	
	IB= 0b00000001;
	out(IB); // instruction byte -> ordre d'ecrire dans le registre CRF2	(24 bits attendus)
		
	out( CFR2[2]);
	out( CFR2[1]);
	out( CFR2[0]);
	
	impulse_IO_update();
}




void AD9951::out_FTW() //Frequency Tuning Word
{

	uint8_t octets[4];

	octets[0] =  FTW & 0b00000000000000000000000011111111;	
	octets[1] = (FTW & 0b00000000000000001111111100000000) >> 8;
	octets[2] = (FTW & 0b00000000111111110000000000000000) >> 16;
	octets[3] = (FTW & 0b11111111000000000000000000000000) >> 24;
	
	IB= 0b00000100;
	out( IB); // instruction byte -> ordre d'écrire 32 bits dans le registre 0x04
	
	out(octets[3]);
	out(octets[2]);
	out(octets[1]);
	out(octets[0]);
	
	impulse_IO_update();
}



void AD9951::out_POW(uint16_t phase_i) // phase offset word
{

	uint8_t octets[2];

	octets[0] =  phase_i & 0b0000000011111111;	
	octets[1] = (phase_i & 0b0011111100000000) >> 8;
	
	IB= 0b00000101;
	out( IB); // instruction byte -> ordre d'écrire 32 bits dans le registre 0x05
	
	out(octets[1]); 
	out(octets[0]);
	
	impulse_IO_update();
}
